import pytest
from typeguard import typechecked

from wyrd.constrained_types import ConstrainedInt, add_constraint


@add_constraint(lambda x: x > 0, "must be at least 1")
class Quantity(ConstrainedInt):
    pass


@typechecked
def total_quantity(a: Quantity, b: Quantity) -> Quantity:
    return Quantity(a + b)


def test_everything_is_fine_when_the_types_are_correct():
    assert total_quantity(Quantity(1), Quantity(2)) == Quantity(3)


def test_error_raised_for_incorrect_types():
    with pytest.raises(TypeError):
        total_quantity(Quantity(1), 2)  # type: ignore



import pytest

from wyrd.constrained_types import ConstrainedString, UnmetConstraintError
from wyrd.constrained_types.helpers import validate


def some_complicated_check(_value):
    raise ValueError("I was always going to fail")


def test_constraint_functions_can_throw_value_errors():
    with pytest.raises(UnmetConstraintError) as err:
        validate(
            ConstrainedString("anything"),
            [(some_complicated_check, "Complicated check")],
        )
    assert str(err.value) == "Complicated check: I was always going to fail"


def test_failing_constraint_is_available_in_exception():
    with pytest.raises(UnmetConstraintError) as err:
        validate(
            ConstrainedString("anything"),
            [(some_complicated_check, "Complicated check")],
        )
    assert err.value.failing_constraint == some_complicated_check



import pytest

from wyrd.constrained_types import (
    UnmetConstraintError,
    ConstrainedFloat,
    add_constraint,
)


@add_constraint(lambda x: x == 3.0, "Pi is exactly 3")
class PiNumber(ConstrainedFloat):
    pass


@add_constraint(lambda x: x >= 1.5, "must be greater than or equal to 1.5")
@add_constraint(lambda x: x <= 10, "must be less than or equal to 9.5")
class OneIshToTenIsh(ConstrainedFloat):
    pass


def test_its_equal_to_a_float():
    assert ConstrainedFloat(3.14159) == 3.14159


def test_it_sums_like_a_float():
    assert ConstrainedFloat(5.0) + ConstrainedFloat(1.5) == ConstrainedFloat(6.5)


@pytest.mark.parametrize(
    "cls,value",
    [(PiNumber, 3), (OneIshToTenIsh, 6.2)],
)
def test_if_a_constraint_is_defined_and_valid_everything_works(cls, value):
    assert cls(value) == value


@pytest.mark.parametrize(
    "cls,value,expected_error",
    [
        (PiNumber, 4, "Pi is exactly 3"),
        (OneIshToTenIsh, 0.3, "must be greater than or equal to 1.5"),
        (OneIshToTenIsh, 10.2, "must be less than or equal to 9.5"),
    ],
)
def test_if_a_constraint_isnt_met_on_construction_an_exception_is_raised(
    cls, value, expected_error
):
    with pytest.raises(UnmetConstraintError) as err:
        cls(value)
    assert str(err.value) == expected_error



from typing import Any, ClassVar, List, Tuple

from .core import Constrained, Constraint
from .helpers import validate


class ConstrainedInt(int, Constrained[int]):
    _constraints: ClassVar[List[Tuple[Constraint[int], str]]] = []

    def __init__(self, value: Any):
        super().__init__()
        self._validate(self)

    @classmethod
    def _validate(cls, value):
        validate(value, cls._constraints)

    # For integration with pydantic
    @classmethod
    def __get_validators__(cls):
        yield lambda v: cls(v)


class ConstrainedString(str, Constrained[str]):
    _constraints: ClassVar[List[Tuple[Constraint[str], str]]] = []

    def __init__(self, value: Any):
        super().__init__()
        self._validate(self)

    @classmethod
    def _validate(cls, value):
        validate(value, cls._constraints)

    # For integration with pydantic
    @classmethod
    def __get_validators__(cls):
        yield lambda v: cls(v)


class ConstrainedFloat(float, Constrained[float]):
    _constraints: ClassVar[List[Tuple[Constraint[float], str]]] = []

    def __init__(self, value: Any):
        super().__init__()
        self._validate(self)

    @classmethod
    def _validate(cls, value):
        validate(value, cls._constraints)

    # For integration with pydantic
    @classmethod
    def __get_validators__(cls):
        yield lambda v: cls(v)



import pytest

from wyrd.constrained_types import UnmetConstraintError, ConstrainedInt, add_constraint


@add_constraint(lambda x: x == 3, "Only 3 is 3")
class NumberThree(ConstrainedInt):
    pass


@add_constraint(lambda x: x >= 1, "must be greater than or equal to 1")
@add_constraint(lambda x: x <= 10, "must be less than or equal to 10")
class OneToTen(ConstrainedInt):
    pass


def test_its_equal_to_an_int():
    assert ConstrainedInt(5) == 5


def test_it_sums_like_an_int():
    assert ConstrainedInt(5) + ConstrainedInt(6) == ConstrainedInt(11)


@pytest.mark.parametrize(
    "cls,value",
    [(NumberThree, 3), (OneToTen, 6)],
)
def test_if_a_constraint_is_defined_and_valid_everything_works(cls, value):
    assert cls(value) == value


@pytest.mark.parametrize(
    "cls,value,expected_error",
    [
        (NumberThree, 4, "Only 3 is 3"),
        (OneToTen, 0, "must be greater than or equal to 1"),
        (OneToTen, 11, "must be less than or equal to 10"),
    ],
)
def test_if_a_constraint_isnt_met_on_construction_an_exception_is_raised(
    cls, value, expected_error
):
    with pytest.raises(UnmetConstraintError) as err:
        cls(value)
    assert str(err.value) == expected_error



import pytest
from pydantic import BaseModel, ValidationError

from wyrd.constrained_types import add_constraint, ConstrainedString, ConstrainedInt


@add_constraint(lambda x: x > 0, "Order must be at least 1")
class OrderQuantity(ConstrainedInt):
    pass


@add_constraint(lambda x: len(x) == 4, "Invalid order id")
class BookId(ConstrainedString):
    pass


class Order(BaseModel):
    book_quantity: OrderQuantity
    book_id: BookId


def test_pydantic_passes_on_validation_failure_as_expected_for_ints():
    with pytest.raises(ValidationError):
        _discount_trick = Order(book_quantity=-1, book_id="ab12")


def test_pydantic_passes_on_validation_failure_as_expected_for_strs():
    with pytest.raises(ValidationError):
        _hackor = Order(book_quantity=2, book_id="drop table lol")


def test_pydantic_accepts_the_extended_types():
    order = Order(book_quantity=OrderQuantity(2), book_id=BookId("abcd"))
    assert isinstance(order, Order)
    assert order.book_quantity == 2
    assert order.book_id == "abcd"



from threading import Lock
from typing import TypeVar, Generic, Type

T = TypeVar("T")


class ReadOnce(Generic[T]):
    has_been_read: bool = False
    _lock: Lock
    __contents: T

    def __init__(self, value: T):
        self.__value = value
        self._lock = Lock()

    def get_contents(self) -> T:
        try:
            self._lock.acquire()
            if self.has_been_read:
                raise RuntimeError(
                    "Attempted to read a ReadOnce value for the second time"
                )
            self.has_been_read = True
            return self.__value
        finally:
            self._lock.release()

    @property
    def type(self) -> Type[T]:
        return type(self.__value)

    def isinstance(self, t: Type) -> bool:
        return isinstance(self.__value, t)



import pytest

from wyrd.constrained_types import (
    add_constraint,
    ConstrainedString,
    UnmetConstraintError,
)


@add_constraint(lambda x: len(x) >= 2, "must be longer than 2 chars")
@add_constraint(lambda x: len(x) >= 1, "must be longer than 1 char")
class SomeString(ConstrainedString):
    pass


def test_the_first_listed_constraint_is_checked_first():
    with pytest.raises(UnmetConstraintError) as err:
        SomeString("")
    assert str(err.value) == "must be longer than 2 chars"



import functools
from typing import TypeVar, ClassVar, List, Tuple, Type, Callable

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


T = TypeVar("T")
U = TypeVar("U", contravariant=True)


class Constraint(Protocol[U]):
    def __call__(self, value: U) -> bool:
        ...


class Constrained(Protocol[T]):
    _constraints: ClassVar[List[Tuple[Constraint[T], str]]]

    @classmethod
    def _validate(cls: Type[T], value: T):
        ...


class UnmetConstraintError(ValueError):
    failing_constraint: Constraint

    def __init__(self, msg: str, failing_constraint: Constraint):
        super().__init__(msg)
        self.failing_constraint = failing_constraint


def add_constraint(
    func: Constraint[T], err_msg: str
) -> Callable[[Type[Constrained[T]]], Type[Constrained[T]]]:
    def decorate(original_class: Type[Constrained[T]]) -> Type[Constrained[T]]:
        _assert_implements_constrained_protocol(original_class)
        new_constraints = [(func, err_msg)] + original_class._constraints

        class NewClass(original_class):  # type: ignore
            _constraints = new_constraints

        NewClass.__name__ = original_class.__name__
        return NewClass

    return decorate


def cache_constraint_results(
    maxsize: int, typed=False
) -> Callable[[Type[Constrained[T]]], Type[Constrained[T]]]:
    def decorate(original_class: Type[Constrained[T]]) -> Type[Constrained[T]]:
        _assert_implements_constrained_protocol(original_class)

        class NewClass(original_class):  # type: ignore
            @classmethod
            @functools.lru_cache(maxsize, typed)
            def _validate(cls, value):
                super()._validate(value)

        NewClass.__name__ = original_class.__name__
        return NewClass

    return decorate


def _assert_implements_constrained_protocol(clas: Type):
    if not (hasattr(clas, "_constraints") and hasattr(clas, "_validate")):
        raise SyntaxError("Class must implement Constrained protocol")



import pytest

from wyrd.constrained_types import (
    ConstrainedString,
    cache_constraint_results,
    add_constraint,
    UnmetConstraintError,
)


checks_made = 0


def some_expensive_check(value):
    global checks_made
    checks_made = checks_made + 1
    return value in ["hello", "bye"]


@add_constraint(some_expensive_check, "must be the string hello or bye")
@cache_constraint_results(maxsize=10)
class SomeString(ConstrainedString):
    pass


def test_check_is_only_made_for_each_value():
    global checks_made
    checks_made = 0
    one = SomeString("hello")
    two = SomeString("bye")
    three = SomeString("hello")

    assert checks_made == 2


def test_check_can_still_fail():
    one = SomeString("hello")
    two = SomeString("bye")

    with pytest.raises(UnmetConstraintError):
        three = SomeString("woooooo!")



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['wyrd', 'wyrd.constrained_types', 'wyrd.read_once']

package_data = \
{'': ['*']}

extras_require = \
{":python_version == '3.6' or python_version == '3.7'": ['typing-extensions']}

setup(name='wyrd',
      version='0.5.0',
      description='Library for helpers with Domain driven security',
      author='meadsteve',
      author_email='meadsteve@gmail.com',
      url='https://github.com/meadsteve/constrained_types/',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
      python_requires='~=3.6',
     )



from typing import List, Tuple, TypeVar

from .core import Constrained, UnmetConstraintError, Constraint

T = TypeVar("T", bound=Constrained)


def validate(value: T, constraints: List[Tuple[Constraint[T], str]]):
    for (is_valid, err_msg) in constraints:
        try:
            valid = is_valid(value)
        except ValueError as e:
            raise UnmetConstraintError(f"{err_msg}: {str(e)}", is_valid) from e
        if not valid:
            raise UnmetConstraintError(err_msg, is_valid)



import pytest

from wyrd.constrained_types import add_constraint, ConstrainedInt


def test_fails_if_you_add_constraints_to_a_plain_class():
    class NormalClass:
        pass

    with pytest.raises(SyntaxError):
        add_constraint(lambda x: True, "always okay")(NormalClass)  # type: ignore


def test_fails_if_you_add_constraints_to_a_plain_class_even_if_it_has_constrains():
    class NormalClass:
        _constraints = []

    with pytest.raises(SyntaxError):
        add_constraint(lambda x: True, "always okay")(NormalClass)  # type: ignore


def test_decorated_class_keeps_its_name():
    @add_constraint(lambda x: True, "always okay")
    class NormalClass(ConstrainedInt):
        pass

    assert NormalClass.__name__ == "NormalClass"



import pytest

from wyrd.constrained_types import (
    UnmetConstraintError,
    ConstrainedString,
    add_constraint,
)


@add_constraint(lambda x: x == "steve", "Only steve is steve")
class Steve(ConstrainedString):
    pass


@add_constraint(lambda x: len(x) >= 1, "must not be empty")
@add_constraint(lambda x: len(x) <= 10, "must not be longer than 10 chars")
class ShortString(ConstrainedString):
    pass


def test_its_equal_to_an_int():
    assert ConstrainedString("boop") == "boop"


def test_it_concats_like_a_string():
    assert ConstrainedString("hello") + ConstrainedString(" you") == ConstrainedString(
        "hello you"
    )


@pytest.mark.parametrize(
    "cls,value",
    [(Steve, "steve"), (ShortString, "okay")],
)
def test_if_a_constraint_is_defined_and_valid_everything_works(cls, value):
    assert cls(value) == value


@pytest.mark.parametrize(
    "cls,value,expected_error",
    [
        (Steve, "stefan", "Only steve is steve"),
        (ShortString, "", "must not be empty"),
        (ShortString, "a" * 11, "must not be longer than 10 chars"),
    ],
)
def test_if_a_constraint_isnt_met_on_construction_an_exception_is_raised(
    cls, value, expected_error
):
    with pytest.raises(UnmetConstraintError) as err:
        cls(value)
    assert str(err.value) == expected_error



"""Extensions for primitives to add validation constraints"""
from .core import UnmetConstraintError, add_constraint, cache_constraint_results
from .primitives import ConstrainedString, ConstrainedInt, ConstrainedFloat

__all__ = [
    "UnmetConstraintError",
    "add_constraint",
    "cache_constraint_results",
    "ConstrainedString",
    "ConstrainedInt",
    "ConstrainedFloat",
]



import pytest

from wyrd.read_once import ReadOnce


def test_it_returns_the_value_stored():
    something = ReadOnce("hello - only once")
    assert something.get_contents() == "hello - only once"


def test_it_raise_an_exception_the_second_time_its_read():
    something = ReadOnce("hello - only once")
    _first = something.get_contents()
    with pytest.raises(RuntimeError):
        _second = something.get_contents()


def test_type_information_can_still_be_accessed():
    assert ReadOnce("hello - only once").type == str
    assert ReadOnce(5).type == int


def test_an_isinstance_equiv_is_provided():
    assert ReadOnce("hello - only once").isinstance(str)
    assert ReadOnce(5).isinstance(object)
